import { defaultArg, value as value_1, some } from "./Option.js";
import { List } from "./Types.js";
import { transpose as transpose_1, pairwise as pairwise_1, collect as collect_1, scanBack as scanBack_1, scan as scan_1, foldBack2 as foldBack2_1, fold2 as fold2_1, fold as fold_1, map as map_1 } from "./Seq.js";
import { getItemFromDict, addToDict, tryGetValue, addToSet, count } from "./Util.js";
import { splitInto as splitInto_1, chunkBySize as chunkBySize_1, permute as permute_1, findIndexBack as findIndexBack_1, tryFindIndexBack as tryFindIndexBack_1, ofList } from "./Array.js";
import { createMutable } from "./Set.js";
import { createMutable as createMutable_1 } from "./Map.js";

export function head(_arg1) {
    if (_arg1.tail != null) {
        return _arg1.head;
    }
    else {
        throw (new Error("List was empty"));
    }
}

export function tryHead(_arg1) {
    if (_arg1.tail != null) {
        return some(_arg1.head);
    }
    else {
        return void 0;
    }
}

export function tail(_arg1) {
    if (_arg1.tail != null) {
        return _arg1.tail;
    }
    else {
        throw (new Error("List was empty"));
    }
}

export function last(_arg1_mut) {
    last:
    while (true) {
        const _arg1 = _arg1_mut;
        if (_arg1.tail != null) {
            if (_arg1.tail.tail == null) {
                return _arg1.head;
            }
            else {
                _arg1_mut = _arg1.tail;
                continue last;
            }
        }
        else {
            throw (new Error("List was empty"));
        }
        break;
    }
}

export function tryLast(_arg1_mut) {
    tryLast:
    while (true) {
        const _arg1 = _arg1_mut;
        if (_arg1.tail != null) {
            if (_arg1.tail.tail == null) {
                return some(_arg1.head);
            }
            else {
                _arg1_mut = _arg1.tail;
                continue tryLast;
            }
        }
        else {
            return void 0;
        }
        break;
    }
}

export function compareWith(comparer, xs, ys) {
    if (xs === ys) {
        return 0;
    }
    else {
        const loop = (xs_1_mut, ys_1_mut) => {
            loop:
            while (true) {
                const xs_1 = xs_1_mut, ys_1 = ys_1_mut;
                if (xs_1.tail != null) {
                    if (ys_1.tail != null) {
                        const matchValue_1 = comparer(xs_1.head, ys_1.head) | 0;
                        if (matchValue_1 === 0) {
                            xs_1_mut = xs_1.tail;
                            ys_1_mut = ys_1.tail;
                            continue loop;
                        }
                        else {
                            return matchValue_1 | 0;
                        }
                    }
                    else {
                        return 1;
                    }
                }
                else if (ys_1.tail == null) {
                    return 0;
                }
                else {
                    return -1;
                }
                break;
            }
        };
        return loop(xs, ys) | 0;
    }
}

export function foldIndexedAux(f_mut, i_mut, acc_mut, _arg1_mut) {
    foldIndexedAux:
    while (true) {
        const f = f_mut, i = i_mut, acc = acc_mut, _arg1 = _arg1_mut;
        if (_arg1.tail != null) {
            f_mut = f;
            i_mut = (i + 1);
            acc_mut = f(i, acc, _arg1.head);
            _arg1_mut = _arg1.tail;
            continue foldIndexedAux;
        }
        else {
            return acc;
        }
        break;
    }
}

export function foldIndexed(f, state, xs) {
    return foldIndexedAux(f, 0, state, xs);
}

export function fold(f_mut, state_mut, xs_mut) {
    fold:
    while (true) {
        const f = f_mut, state = state_mut, xs = xs_mut;
        if (xs.tail != null) {
            f_mut = f;
            state_mut = f(state, xs.head);
            xs_mut = xs.tail;
            continue fold;
        }
        else {
            return state;
        }
        break;
    }
}

export function reverse(xs) {
    return fold((acc, x) => (new List(x, acc)), new List(), xs);
}

export function foldBack(f, xs, state) {
    return fold((acc, x) => f(x, acc), state, reverse(xs));
}

export function toSeq(xs) {
    return map_1((x) => x, xs);
}

export function ofSeq(xs) {
    const xs_1 = fold_1((acc, x) => (new List(x, acc)), new List(), xs);
    return reverse(xs_1);
}

export function concat(lists) {
    const xs_1 = fold_1((state, xs) => fold((acc, x) => (new List(x, acc)), state, xs), new List(), lists);
    return reverse(xs_1);
}

export function foldIndexed2Aux(f_mut, i_mut, acc_mut, bs_mut, cs_mut) {
    foldIndexed2Aux:
    while (true) {
        const f = f_mut, i = i_mut, acc = acc_mut, bs = bs_mut, cs = cs_mut;
        let pattern_matching_result, x, xs, y, ys;
        if (bs.tail != null) {
            if (cs.tail != null) {
                pattern_matching_result = 1;
                x = bs.head;
                xs = bs.tail;
                y = cs.head;
                ys = cs.tail;
            }
            else {
                pattern_matching_result = 2;
            }
        }
        else if (cs.tail == null) {
            pattern_matching_result = 0;
        }
        else {
            pattern_matching_result = 2;
        }
        switch (pattern_matching_result) {
            case 0: {
                return acc;
            }
            case 1: {
                f_mut = f;
                i_mut = (i + 1);
                acc_mut = f(i, acc, x, y);
                bs_mut = xs;
                cs_mut = ys;
                continue foldIndexed2Aux;
            }
            case 2: {
                throw (new Error("Lists had different lengths"));
            }
        }
        break;
    }
}

export function foldIndexed2(f, state, xs, ys) {
    return foldIndexed2Aux(f, 0, state, xs, ys);
}

export function fold2(f, state, xs, ys) {
    return fold2_1(f, state, xs, ys);
}

export function foldBack2(f, xs, ys, state) {
    return foldBack2_1(f, xs, ys, state);
}

export function unfold(f, state) {
    const unfoldInner = (acc_mut, state_1_mut) => {
        unfoldInner:
        while (true) {
            const acc = acc_mut, state_1 = state_1_mut;
            const matchValue = f(state_1);
            if (matchValue != null) {
                const x = matchValue[0];
                const state_2 = matchValue[1];
                acc_mut = (new List(x, acc));
                state_1_mut = state_2;
                continue unfoldInner;
            }
            else {
                return reverse(acc);
            }
            break;
        }
    };
    return unfoldInner(new List(), state);
}

export function foldIndexed3Aux(f_mut, i_mut, acc_mut, bs_mut, cs_mut, ds_mut) {
    foldIndexed3Aux:
    while (true) {
        const f = f_mut, i = i_mut, acc = acc_mut, bs = bs_mut, cs = cs_mut, ds = ds_mut;
        let pattern_matching_result, x, xs, y, ys, z, zs;
        if (bs.tail != null) {
            if (cs.tail != null) {
                if (ds.tail != null) {
                    pattern_matching_result = 1;
                    x = bs.head;
                    xs = bs.tail;
                    y = cs.head;
                    ys = cs.tail;
                    z = ds.head;
                    zs = ds.tail;
                }
                else {
                    pattern_matching_result = 2;
                }
            }
            else {
                pattern_matching_result = 2;
            }
        }
        else if (cs.tail == null) {
            if (ds.tail == null) {
                pattern_matching_result = 0;
            }
            else {
                pattern_matching_result = 2;
            }
        }
        else {
            pattern_matching_result = 2;
        }
        switch (pattern_matching_result) {
            case 0: {
                return acc;
            }
            case 1: {
                f_mut = f;
                i_mut = (i + 1);
                acc_mut = f(i, acc, x, y, z);
                bs_mut = xs;
                cs_mut = ys;
                ds_mut = zs;
                continue foldIndexed3Aux;
            }
            case 2: {
                throw (new Error("Lists had different lengths"));
            }
        }
        break;
    }
}

export function foldIndexed3(f, seed, xs, ys, zs) {
    return foldIndexed3Aux(f, 0, seed, xs, ys, zs);
}

export function fold3(f, state, xs, ys, zs) {
    return foldIndexed3((_arg1, acc, x, y, z) => f(acc, x, y, z), state, xs, ys, zs);
}

export function scan(f, state, xs) {
    const xs_1 = scan_1(f, state, xs);
    return ofSeq(xs_1);
}

export function scanBack(f, xs, state) {
    const xs_1 = scanBack_1(f, xs, state);
    return ofSeq(xs_1);
}

export function length(xs) {
    return fold((acc, _arg1) => (acc + 1), 0, xs);
}

export function append(xs, ys) {
    return fold((acc, x) => (new List(x, acc)), ys, reverse(xs));
}

export function collect(f, xs) {
    const xs_1 = collect_1(f, xs);
    return ofSeq(xs_1);
}

export function map(f, xs) {
    const xs_1 = fold((acc, x) => (new List(f(x), acc)), new List(), xs);
    return reverse(xs_1);
}

export function mapIndexed(f, xs) {
    const xs_1 = foldIndexed((i, acc, x) => (new List(f(i, x), acc)), new List(), xs);
    return reverse(xs_1);
}

export function indexed(xs) {
    return mapIndexed((i, x) => ([i, x]), xs);
}

export function map2(f, xs, ys) {
    const xs_1 = fold2((acc, x, y) => (new List(f(x, y), acc)), new List(), xs, ys);
    return reverse(xs_1);
}

export function mapIndexed2(f, xs, ys) {
    const xs_1 = foldIndexed2((i, acc, x, y) => (new List(f(i, x, y), acc)), new List(), xs, ys);
    return reverse(xs_1);
}

export function map3(f, xs, ys, zs) {
    const xs_1 = fold3((acc, x, y, z) => (new List(f(x, y, z), acc)), new List(), xs, ys, zs);
    return reverse(xs_1);
}

export function mapIndexed3(f, xs, ys, zs) {
    const xs_1 = foldIndexed3((i, acc, x, y, z) => (new List(f(i, x, y, z), acc)), new List(), xs, ys, zs);
    return reverse(xs_1);
}

export function mapFold(f, s, xs) {
    const patternInput_1 = fold((tupledArg, x) => {
        const patternInput = f(tupledArg[1], x);
        return [new List(patternInput[0], tupledArg[0]), patternInput[1]];
    }, [new List(), s], xs);
    return [reverse(patternInput_1[0]), patternInput_1[1]];
}

export function mapFoldBack(f, xs, s) {
    return mapFold((s_1, v) => f(v, s_1), s, reverse(xs));
}

export function iterate(f, xs) {
    fold((unitVar0, x) => {
        f(x);
    }, void 0, xs);
}

export function iterate2(f, xs, ys) {
    fold2((unitVar0, x, y) => {
        f(x, y);
    }, void 0, xs, ys);
}

export function iterateIndexed(f, xs) {
    foldIndexed((i, unitVar1, x) => {
        f(i, x);
    }, void 0, xs);
}

export function iterateIndexed2(f, xs, ys) {
    foldIndexed2((i, unitVar1, x, y) => {
        f(i, x, y);
    }, void 0, xs, ys);
}

export function ofArray(xs) {
    let res = new List();
    for (let i = count(xs) - 1; i >= 0; i--) {
        res = (new List(xs[i], res));
    }
    return res;
}

export function empty() {
    return new List();
}

export function isEmpty(_arg1) {
    if (_arg1.tail == null) {
        return true;
    }
    else {
        return false;
    }
}

export function tryPickIndexedAux(f_mut, i_mut, _arg1_mut) {
    tryPickIndexedAux:
    while (true) {
        const f = f_mut, i = i_mut, _arg1 = _arg1_mut;
        if (_arg1.tail != null) {
            const result = f(i, _arg1.head);
            if (result == null) {
                f_mut = f;
                i_mut = (i + 1);
                _arg1_mut = _arg1.tail;
                continue tryPickIndexedAux;
            }
            else {
                return result;
            }
        }
        else {
            return void 0;
        }
        break;
    }
}

export function tryPickIndexed(f, xs) {
    return tryPickIndexedAux(f, 0, xs);
}

export function tryPick(f, xs) {
    return tryPickIndexed((_arg1, x) => f(x), xs);
}

export function pick(f, xs) {
    const matchValue = tryPick(f, xs);
    if (matchValue != null) {
        const x = value_1(matchValue);
        return x;
    }
    else {
        throw (new Error("List did not contain any matching elements"));
    }
}

export function tryFindIndexed(f, xs) {
    return tryPickIndexed((i, x) => (f(i, x) ? some(x) : (void 0)), xs);
}

export function tryFind(f, xs) {
    return tryPickIndexed((_arg1, x) => (f(x) ? some(x) : (void 0)), xs);
}

export function findIndexed(f, xs) {
    const matchValue = tryFindIndexed(f, xs);
    if (matchValue != null) {
        const x = value_1(matchValue);
        return x;
    }
    else {
        throw (new Error("List did not contain any matching elements"));
    }
}

export function find(f, xs) {
    return findIndexed((_arg1, x) => f(x), xs);
}

export function findBack(f, xs) {
    let xs_2;
    xs_2 = reverse(xs);
    return find(f, xs_2);
}

export function tryFindBack(f, xs) {
    let xs_2;
    xs_2 = reverse(xs);
    return tryFind(f, xs_2);
}

export function tryFindIndex(f, xs) {
    return tryPickIndexed((i, x) => (f(x) ? i : (void 0)), xs);
}

export function tryFindIndexBack(f, xs) {
    const array = ofList(xs, Array);
    return tryFindIndexBack_1(f, array);
}

export function findIndex(f, xs) {
    const matchValue = tryFindIndex(f, xs);
    if (matchValue != null) {
        const x = matchValue | 0;
        return x | 0;
    }
    else {
        throw (new Error("List did not contain any matching elements"));
    }
}

export function findIndexBack(f, xs) {
    const array = ofList(xs, Array);
    return findIndexBack_1(f, array) | 0;
}

export function item(n, xs) {
    return findIndexed((i, _arg1) => (n === i), xs);
}

export function tryItem(n, xs) {
    return tryFindIndexed((i, _arg1) => (n === i), xs);
}

export function filter(f, xs) {
    const xs_1 = fold((acc, x) => (f(x) ? (new List(x, acc)) : acc), new List(), xs);
    return reverse(xs_1);
}

export function partition(f, xs) {
    return fold((tupledArg, x) => (f(x) ? ([new List(x, tupledArg[0]), tupledArg[1]]) : ([tupledArg[0], new List(x, tupledArg[1])])), [new List(), new List()], reverse(xs));
}

export function choose(f, xs) {
    const xs_1 = fold((acc, x) => {
        const matchValue = f(x);
        if (matchValue == null) {
            return acc;
        }
        else {
            const y = value_1(matchValue);
            return new List(y, acc);
        }
    }, new List(), xs);
    return reverse(xs_1);
}

export function contains(value, list, eq) {
    const loop = (xs_mut) => {
        loop:
        while (true) {
            const xs = xs_mut;
            if (xs.tail != null) {
                if (eq.Equals(value, xs.head)) {
                    return true;
                }
                else {
                    xs_mut = xs.tail;
                    continue loop;
                }
            }
            else {
                return false;
            }
            break;
        }
    };
    return loop(list);
}

export function except(itemsToExclude, array, eq) {
    if (isEmpty(array)) {
        return array;
    }
    else {
        const cached = createMutable(itemsToExclude, eq);
        return filter((arg00) => addToSet(arg00, cached), array);
    }
}

export function initialize(n, f) {
    let xs = new List();
    for (let i = 0; i <= (n - 1); i++) {
        xs = (new List(f(i), xs));
    }
    return reverse(xs);
}

export function replicate(n, x) {
    return initialize(n, (_arg1) => x);
}

export function reduce(f, _arg1) {
    if (_arg1.tail != null) {
        return fold(f, _arg1.head, _arg1.tail);
    }
    else {
        throw (new Error("List was empty"));
    }
}

export function reduceBack(f, _arg1) {
    if (_arg1.tail != null) {
        return foldBack(f, _arg1.tail, _arg1.head);
    }
    else {
        throw (new Error("List was empty"));
    }
}

export function forAll(f, xs) {
    return fold((acc, x) => (acc ? f(x) : false), true, xs);
}

export function forAll2(f, xs, ys) {
    return fold2((acc, x, y) => (acc ? f(x, y) : false), true, xs, ys);
}

export function exists(f_mut, _arg1_mut) {
    exists:
    while (true) {
        const f = f_mut, _arg1 = _arg1_mut;
        if (_arg1.tail != null) {
            if (f(_arg1.head)) {
                return true;
            }
            else {
                f_mut = f;
                _arg1_mut = _arg1.tail;
                continue exists;
            }
        }
        else {
            return false;
        }
        break;
    }
}

export function exists2(f_mut, bs_mut, cs_mut) {
    exists2:
    while (true) {
        const f = f_mut, bs = bs_mut, cs = cs_mut;
        let pattern_matching_result, x, xs, y, ys;
        if (bs.tail != null) {
            if (cs.tail != null) {
                pattern_matching_result = 1;
                x = bs.head;
                xs = bs.tail;
                y = cs.head;
                ys = cs.tail;
            }
            else {
                pattern_matching_result = 2;
            }
        }
        else if (cs.tail == null) {
            pattern_matching_result = 0;
        }
        else {
            pattern_matching_result = 2;
        }
        switch (pattern_matching_result) {
            case 0: {
                return false;
            }
            case 1: {
                if (f(x, y)) {
                    return true;
                }
                else {
                    f_mut = f;
                    bs_mut = xs;
                    cs_mut = ys;
                    continue exists2;
                }
            }
            case 2: {
                throw (new Error("Lists had different lengths"));
            }
        }
        break;
    }
}

export function unzip(xs) {
    return foldBack((tupledArg, tupledArg_1) => ([new List(tupledArg[0], tupledArg_1[0]), new List(tupledArg[1], tupledArg_1[1])]), xs, [new List(), new List()]);
}

export function unzip3(xs) {
    return foldBack((tupledArg, tupledArg_1) => ([new List(tupledArg[0], tupledArg_1[0]), new List(tupledArg[1], tupledArg_1[1]), new List(tupledArg[2], tupledArg_1[2])]), xs, [new List(), new List(), new List()]);
}

export function zip(xs, ys) {
    return map2((x, y) => ([x, y]), xs, ys);
}

export function zip3(xs, ys, zs) {
    return map3((x, y, z) => ([x, y, z]), xs, ys, zs);
}

export function sort(xs, comparer) {
    let xs_2;
    const xs_1 = ofList(xs, Array);
    xs_1.sort(((x, y) => comparer.Compare(x, y)));
    xs_2 = xs_1;
    return ofArray(xs_2);
}

export function sortBy(projection, xs, comparer) {
    let xs_2;
    const xs_1 = ofList(xs, Array);
    xs_1.sort(((x, y) => comparer.Compare(projection(x), projection(y))));
    xs_2 = xs_1;
    return ofArray(xs_2);
}

export function sortDescending(xs, comparer) {
    let xs_2;
    const xs_1 = ofList(xs, Array);
    xs_1.sort(((x, y) => (comparer.Compare(x, y) * -1)));
    xs_2 = xs_1;
    return ofArray(xs_2);
}

export function sortByDescending(projection, xs, comparer) {
    let xs_2;
    const xs_1 = ofList(xs, Array);
    xs_1.sort(((x, y) => (comparer.Compare(projection(x), projection(y)) * -1)));
    xs_2 = xs_1;
    return ofArray(xs_2);
}

export function sortWith(comparer, xs) {
    let xs_2;
    const comparer_1 = comparer;
    const xs_1 = ofList(xs, Array);
    xs_1.sort(comparer_1);
    xs_2 = xs_1;
    return ofArray(xs_2);
}

export function sum(xs, adder) {
    return fold((acc, x) => adder.Add(acc, x), adder.GetZero(), xs);
}

export function sumBy(f, xs, adder) {
    return fold((acc, x) => adder.Add(acc, f(x)), adder.GetZero(), xs);
}

export function maxBy(projection, xs, comparer) {
    return reduce((x, y) => ((comparer.Compare(projection(y), projection(x)) > 0) ? y : x), xs);
}

export function max(li, comparer) {
    return reduce((x, y) => ((comparer.Compare(y, x) > 0) ? y : x), li);
}

export function minBy(projection, xs, comparer) {
    return reduce((x, y) => ((comparer.Compare(projection(y), projection(x)) > 0) ? x : y), xs);
}

export function min(xs, comparer) {
    return reduce((x, y) => ((comparer.Compare(y, x) > 0) ? x : y), xs);
}

export function average(xs, averager) {
    const total = fold((acc, x) => averager.Add(acc, x), averager.GetZero(), xs);
    return averager.DivideByInt(total, length(xs));
}

export function averageBy(f, xs, averager) {
    const total = fold((acc, x) => averager.Add(acc, f(x)), averager.GetZero(), xs);
    return averager.DivideByInt(total, length(xs));
}

export function permute(f, xs) {
    let xs_1;
    let array;
    array = ofList(xs, Array);
    xs_1 = permute_1(f, array);
    return ofArray(xs_1);
}

export function chunkBySize(chunkSize, xs) {
    let xs_3;
    let xs_1;
    let array;
    array = ofList(xs, Array);
    xs_1 = chunkBySize_1(chunkSize, array);
    xs_3 = ofArray(xs_1);
    return map(ofArray, xs_3);
}

export function skip(i, xs) {
    const skipInner = (i_1_mut, xs_1_mut) => {
        skipInner:
        while (true) {
            const i_1 = i_1_mut, xs_1 = xs_1_mut;
            if (i_1 === 0) {
                return xs_1;
            }
            else if (xs_1.tail != null) {
                i_1_mut = (i_1 - 1);
                xs_1_mut = xs_1.tail;
                continue skipInner;
            }
            else {
                throw (new Error("The input sequence has an insufficient number of elements."));
            }
            break;
        }
    };
    if (i < 0) {
        throw (new Error("The input must be non-negative."));
    }
    else {
        let pattern_matching_result, i_4, xs_4;
        if (i === 0) {
            pattern_matching_result = 0;
        }
        else if (i === 1) {
            if (xs.tail != null) {
                pattern_matching_result = 1;
            }
            else {
                pattern_matching_result = 2;
                i_4 = i;
                xs_4 = xs;
            }
        }
        else {
            pattern_matching_result = 2;
            i_4 = i;
            xs_4 = xs;
        }
        switch (pattern_matching_result) {
            case 0: {
                return xs;
            }
            case 1: {
                return xs.tail;
            }
            case 2: {
                return skipInner(i_4, xs_4);
            }
        }
    }
}

export function skipWhile(predicate_mut, xs_mut) {
    skipWhile:
    while (true) {
        const predicate = predicate_mut, xs = xs_mut;
        let pattern_matching_result, h_1, t_1;
        if (xs.tail != null) {
            if (predicate(xs.head)) {
                pattern_matching_result = 0;
                h_1 = xs.head;
                t_1 = xs.tail;
            }
            else {
                pattern_matching_result = 1;
            }
        }
        else {
            pattern_matching_result = 1;
        }
        switch (pattern_matching_result) {
            case 0: {
                predicate_mut = predicate;
                xs_mut = t_1;
                continue skipWhile;
            }
            case 1: {
                return xs;
            }
        }
        break;
    }
}

export function takeSplitAux(error_mut, i_mut, acc_mut, xs_mut) {
    takeSplitAux:
    while (true) {
        const error = error_mut, i = i_mut, acc = acc_mut, xs = xs_mut;
        if (i === 0) {
            return [reverse(acc), xs];
        }
        else if (xs.tail != null) {
            error_mut = error;
            i_mut = (i - 1);
            acc_mut = (new List(xs.head, acc));
            xs_mut = xs.tail;
            continue takeSplitAux;
        }
        else {
            if (error) {
                throw (new Error("The input sequence has an insufficient number of elements."));
            }
            else {
                return [reverse(acc), xs];
            }
        }
        break;
    }
}

export function take(i, xs) {
    if (i < 0) {
        throw (new Error("The input must be non-negative."));
    }
    else {
        let pattern_matching_result, i_3, xs_1;
        if (i === 0) {
            pattern_matching_result = 0;
        }
        else if (i === 1) {
            if (xs.tail != null) {
                pattern_matching_result = 1;
            }
            else {
                pattern_matching_result = 2;
                i_3 = i;
                xs_1 = xs;
            }
        }
        else {
            pattern_matching_result = 2;
            i_3 = i;
            xs_1 = xs;
        }
        switch (pattern_matching_result) {
            case 0: {
                return new List();
            }
            case 1: {
                return new List(xs.head, new List());
            }
            case 2: {
                const tuple = takeSplitAux(true, i_3, new List(), xs_1);
                return tuple[0];
            }
        }
    }
}

export function takeWhile(predicate, xs) {
    if (xs.tail != null) {
        if (xs.tail.tail == null) {
            if (predicate(xs.head)) {
                return xs;
            }
            else {
                return xs.tail;
            }
        }
        else {
            if (!predicate(xs.head)) {
                return new List();
            }
            else {
                return new List(xs.head, takeWhile(predicate, xs.tail));
            }
        }
    }
    else {
        return xs;
    }
}

export function truncate(i, xs) {
    if (i < 0) {
        throw (new Error("The input must be non-negative."));
    }
    else {
        let pattern_matching_result, i_3, xs_1;
        if (i === 0) {
            pattern_matching_result = 0;
        }
        else if (i === 1) {
            if (xs.tail != null) {
                pattern_matching_result = 1;
            }
            else {
                pattern_matching_result = 2;
                i_3 = i;
                xs_1 = xs;
            }
        }
        else {
            pattern_matching_result = 2;
            i_3 = i;
            xs_1 = xs;
        }
        switch (pattern_matching_result) {
            case 0: {
                return new List();
            }
            case 1: {
                return new List(xs.head, new List());
            }
            case 2: {
                const tuple = takeSplitAux(false, i_3, new List(), xs_1);
                return tuple[0];
            }
        }
    }
}

export function splitAt(i, xs) {
    if (i < 0) {
        throw (new Error("The input must be non-negative."));
    }
    else {
        let pattern_matching_result, i_3, xs_2;
        if (i === 0) {
            pattern_matching_result = 0;
        }
        else if (i === 1) {
            if (xs.tail != null) {
                pattern_matching_result = 1;
            }
            else {
                pattern_matching_result = 2;
                i_3 = i;
                xs_2 = xs;
            }
        }
        else {
            pattern_matching_result = 2;
            i_3 = i;
            xs_2 = xs;
        }
        switch (pattern_matching_result) {
            case 0: {
                return [new List(), xs];
            }
            case 1: {
                return [new List(xs.head, new List()), xs.tail];
            }
            case 2: {
                return takeSplitAux(true, i_3, new List(), xs_2);
            }
        }
    }
}

export function outOfRange() {
    throw (new Error("Index out of range"));
}

export function slice(lower, upper, xs) {
    const lower_1 = defaultArg(lower, 0) | 0;
    const hasUpper = upper != null;
    if (lower_1 < 0) {
        return outOfRange();
    }
    else if (hasUpper ? (upper < lower_1) : false) {
        return new List();
    }
    else {
        let lastIndex = -1;
        let res;
        const state = new List();
        res = foldIndexed((i, acc, x) => {
            lastIndex = i;
            if ((lower_1 <= i) ? ((!hasUpper) ? true : (i <= upper)) : false) {
                return new List(x, acc);
            }
            else {
                return acc;
            }
        }, state, xs);
        if ((lower_1 > (lastIndex + 1)) ? true : (hasUpper ? (upper > lastIndex) : false)) {
            outOfRange();
        }
        else {
        }
        return reverse(res);
    }
}

export function distinctBy(projection, xs, eq) {
    const hashSet = createMutable([], eq);
    return filter((arg) => {
        const arg00 = projection(arg);
        return addToSet(arg00, hashSet);
    }, xs);
}

export function distinct(xs, eq) {
    return distinctBy((x) => x, xs, eq);
}

export function exactlyOne(xs) {
    if (xs.tail != null) {
        if (xs.tail.tail != null) {
            throw (new Error("Input list too long\\nParameter name: list"));
        }
        else {
            return xs.head;
        }
    }
    else {
        throw (new Error("The input sequence was empty\\nParameter name: list"));
    }
}

export function groupBy(projection, xs, eq) {
    const dict = createMutable_1([], eq);
    let keys = new List();
    iterate((v) => {
        const key = projection(v);
        const matchValue = tryGetValue(dict, key, null);
        if (matchValue[0]) {
            dict.set(key, new List(v, matchValue[1]));
        }
        else {
            addToDict(dict, key, new List(v, new List()));
            keys = (new List(key, keys));
        }
    }, xs);
    let result = new List();
    const xs_2 = keys;
    iterate((key_1) => {
        result = (new List([key_1, reverse(getItemFromDict(dict, key_1))], result));
    }, xs_2);
    return result;
}

export function countBy(projection, xs, eq) {
    const dict = createMutable_1([], eq);
    let keys = new List();
    iterate((v) => {
        const key = projection(v);
        const matchValue = tryGetValue(dict, key, 0);
        if (matchValue[0]) {
            dict.set(key, matchValue[1] + 1);
        }
        else {
            dict.set(key, 1);
            keys = (new List(key, keys));
        }
    }, xs);
    let result = new List();
    const xs_2 = keys;
    iterate((key_1) => {
        result = (new List([key_1, getItemFromDict(dict, key_1)], result));
    }, xs_2);
    return result;
}

export function where(predicate, source) {
    return filter(predicate, source);
}

export function pairwise(source) {
    const xs = pairwise_1(source);
    return ofSeq(xs);
}

export function windowed(windowSize, source) {
    if (windowSize <= 0) {
        throw (new Error("windowSize must be positive"));
    }
    else {
    }
    let res = new List();
    for (let i = length(source); i >= windowSize; i--) {
        res = (new List(slice(i - windowSize, i - 1, source), res));
    }
    return res;
}

export function splitInto(chunks, source) {
    let xs_2;
    let xs;
    let array;
    array = ofList(source, Array);
    xs = splitInto_1(chunks, array);
    xs_2 = ofArray(xs);
    return map(ofArray, xs_2);
}

export function transpose(lists) {
    let xs_1;
    let source_1;
    source_1 = transpose_1(lists);
    xs_1 = map_1(ofSeq, source_1);
    return ofSeq(xs_1);
}

